{% extends '!footer.html' %}
{% block extrafooter %}
<head>
<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css">
</head>
<br></br>
<i style="font-size:24px" class="fa">&#xf09b;</i>
<strong><u>GitHub</u>:</strong> <a href="https://github.com/bbugyi200/clap">https://github.com/bbugyi200/clap</a>
{% endblock %}
